/*****************************************************************************
 *                                  M A R Q S R 1
 *****************************************************************************
 *
 *   PROGRAM ID:        MARQSR1.C
 *
 *   AUTHOR:            Glynne Casteel
 *                      GlynneC@ix.netcom.com
 *
 *                      This software IS copyrighted, BUT you may use it freely
 *                      PROVIDED THAT you send me a copy of any commercial or
 *                      shareware product that incorporates this code.
 *
 *
 *   DATE:              June 11, 1994
 *
 *   DESCRIPTION:
 *
 *      The Levenberg-Marquardt-Broyden NLE solver was successful, so I
 *      decided to try a Levenberg-Marquardt-Optimization method.  It
 *      uses a unique strategy to switch between the (Quasi-)Newton and
 *      the Gradient direction, which requires no line searches.
 *
 *      My modifications to the algorithm are as follows:
 *      Instead of solving  (J~.J + uI).dx = -J~.f
 *                    for   dx at each step,
 *
 *      I picked a ration polynomial function of u that retains the
 *      asymptotic dx-values at u=0 and  u=infinity.
 *
 *          dx= -(ug + H.f)/(1+uu)
 *          at u=0,   dx= -H.g ---> Newton-Raphson step
 *          at u=inf, dx= -g/u ---> tiny Gradient step
 *
 *      My second trick is to use the Symmetric Rank 1 (SR1) method
 *      to update the inverse Jacobian (H = I/J):
 *
 *          H= H +(dx-H.dg)(dx-H.dg)/(dx.dg-dg.H.dg)
 *
 *
 *
 *   INPUT PARAMETERS:  None
 *
 *   RETURN/EXIT VALUE: Number of iterations required for solution
 *
 *   INPUT FILES:       None
 *
 *   OUTPUT FILES:      None
 *
 *   COMPILE/LINK:      Microsoft C 6.0 compatable compiler
 *
 *
 *   SPECIAL NOTES:     None
 *
 *****************************************************************************
 *                           MODIFICATION LOG
 *
 *   DATE          NAME                DESCRIPTION
 *   ------------  ------------------  ----------------------------------
 *
 ******************************************************************************/



/*
   When the COLLINS_EXTERNS flag is turned on it puts the following
   types of definitions into the source:

    extern unsigned long enStep;

   In the application the flag isn't on so declaration occurs.
   These variables are used to communicate progress of a subroutine
   to the application.
*/
#define COLLINS_EXTERNS   1
#include <collins.h>

// locally global variables
static OPTFXN F;
static GRADF grad;
static int nLoops;


// local functions and macros
#define TINY   1.0e-10
static void opt( int n, float x[], float tol, float g[], float s[],
                 float dg[], float Hdg[], float Hg[], float dx[], float *H[] );
static void gradF( int n, float x[], float F0, float g[] );



/*
   ******* HERE IS THE ENTRY POINT INTO THIS MODULE *********
*/
int marqsr1( int n, float x[], OPTFXN FXN, GRADF DFXN, float tol )
/*
   This subroutine is just a dummy to allocate the n(n+6)
   dimensional workspace w which is needed by the following
   subroutines. This is a holdover from the FORTRAN roots of
   this library. In FORTRAN, you can't dynamically allocate
   an array, so the calling program must create a "work array"
   of a certain size. Then you use clever function parameter
   mapping to divvy up this array into all of the vectors and
   matrices needed in the library module.

   This subroutine also assigns user defined functions to
   (locally) global pointers, so we don't have to keep passing
   them to each lower level.
*/
{
    float **w;
    w= matrix( 1, n+6, 1, n );
    F= FXN;
    if( DFXN )
        grad= DFXN;
    else
        grad= gradF;

    if ( tol <= 0.0 )  tol= 1.0e-6;
///****/ printf( "marqbfgs Entry: x= (%e %e)\n", x[1], x[2] );
    opt( n, x, tol, w[1], w[2], w[3], w[4], w[5], w[6], &w[6] );
    free_matrix( w, 1, n+6, 1, n );

    return( nLoops );
}




static void opt( int n, float x[], float tol, float g[], float s[],
                 float dg[], float Hdg[], float Hg[], float dx[], float *H[] )

/* ----------------------VARIABLE KEY-----------------------------
   n          length of independent (vector) variable

   x          independent variable-- on input it contains
              the initial guess;  on output, the solution

   F          objective function (to be supplied by user!!)

   F0..F3     stores previous values of F

   g          stores value of gradF call

   s          the search direction

   dg         the change in g between steps

   t          scale factor applied to s

   H          "iteration matrix" --approx inverse of the Jacobian

   Hg, Hdg    the vector dot products   H.g, H.dg

   sg, sdg,   the scalar dot products   s.g, s.dg,
   dgHdg,                               dg.H.dg, gHdg
   gHdg

------------------------------------------------------------------
*/
{
int i, k, nMain, bRestart, bConverged, goodloops;
float F0, F1, F2, F3, sg, sdg, gabs, sabs, dgHdg, gHdg, u;


// initializations
    F1= 0.0;
    F0= F( n, x );
    grad( n, x, F0, g );
    nLoops= 0;
    nMain= 0;
    bRestart= 1;
    bConverged= 0;


// main loop
    while (  !bConverged &&  nMain<50  )
    {
        printf( "%5d  F=%16.8e\r", nLoops, F0 );

        if ( bRestart )
        {
            bRestart= 0;
            nMain++;
            goodloops=0;
            u= n * 1000;

            gabs= sabs= 0;
            for ( k=1 ; k<=n ; k++ )
            {
                for ( i=k ; i<=n ; i++ )
                    H[i][k]= H[k][i]= 0.0;
                H[k][k]= 1.0;
                s[k]= -g[k]/u;
                dg[k]= g[k];
                gabs += g[k]*g[k];
                sabs += s[k]*s[k];
            }
            sabs= sqrt( sabs );
            gabs= sqrt( gabs );
        }


     // calc Marquardt correction; the way I wrote the initial SR1,
     // +s is the newton direction, there is a difference of sign
     // between this and the Marquardt-Broyden module.
///****/ printf( "marqbfgs: u= %e, F= %e\n", u, F0 );
///****/ printf( "\t  g= (%e %e)\n", g[1],  g[2] );
///****/ printf( "\t  x= (%e %e)\n", x[1],  x[2] );
///****/ printf( "\t  s= (%e %e)\n", s[1],  s[2] );
        for( k=1 ; k<=n ; k++ )
        {
            dx[k]= ( s[k] -sabs/gabs*u*g[k]) / (u*u +1);
            x[k] += dx[k];

        }
///****/ printf( "\t dx= (%e %e)\n",dx[1],dx[2] );


        // obtain new function values
        F1= F0;
        F0= F( n, x );



// *** Exit test ***
        if ( fabs(F1-F0) <= 0.5*tol*(fabs(F1)+fabs(F0)+TINY) )
            bConverged= 1;
        else if( F0 > F1 )
        {
            // redo last step with larger u-value
///            u *= 5;   //notice that I don't increase on a bad step as much as
                      //I decrease on a good step. This ensures we drift to
                      //the Newton direction over time.
            u *= 10;
            for( k=1 ; k<=n ; k++ )
            {
                x[k] -= dx[k];
                F0= F1;
            }
        }
        else
        {
            goodloops++;

            // shorten u-value; calc new g, H, s
            u *= 0.1;
///            u *= 0.5;

            grad( n, x, F0, g );
            for ( i=1 ; i<=n ; i++ )
            {
                dg[i]= g[i] -dg[i];  // true dg calculation
                s[i]= dx[i];  // here's the actual step for update calculations
            }

            // calculate dot products
            sg= sdg= gHdg= dgHdg= 0.0;
            for ( i=1 ; i<=n ; i++ )
            {
                Hg[i] = 0.0;
                Hdg[i]= 0.0;
                for ( k=1 ; k<=n ; k++ )
                {
                    Hg[i]  += H[k][i]*g[k];
                    Hdg[i] += H[k][i]*dg[k];
                }
                sg  += s[i]*g[i];
                sdg += s[i]*dg[i];
                gHdg  += g[i]*Hdg[i];
                dgHdg += dg[i]*Hdg[i];
            }
///****/ printf( "\t Hg= (%e %e)\n", Hg[1], Hg[2] );
///****/ printf( "\t Hdg= (%e %e)\n", Hdg[1], Hdg[2] );
///****/ printf( "\t  sg= %e, sdg= %e, gHdg= %e, dgHdg= %e\n", sg,sdg,gHdg,dgHdg );
///****/ printf( "\t H= (%e %e)\n\t    (%e %e)\n", H[1][1], H[1][2],H[2][1],H[2][2]  );


          //  update H and calc new s utilizing the symmetry of H and
          //  the long form of H-update to avoid extra loop for s= -H.g

            gabs= sabs= 0;
            for ( i=1 ; i<=n ; i++ )
            {
                for ( k=i ; k<=n ; k++ )
                {
                    H[k][i] += (s[k]-Hdg[k]) * (s[i]-Hdg[i]) / (sdg-dgHdg);
                    H[i][k]= H[k][i];
                }
                s[i]= -Hg[i] -(s[i]-Hdg[i])*(sg-gHdg)/(sdg-dgHdg);

                dg[i]= g[i];  // dump off old g-value

                gabs += g[i]*g[i];
                sabs += s[i]*s[i];
            }
            sabs= sqrt( sabs );
            gabs= sqrt( gabs );

        }

        // *** Restart Test ***
        nLoops++;
        if( nLoops%(20*n*n) == 0 )
            bRestart= 1;
        else if( goodloops%(2*n*n) == 0   &&   u > 0.01  )
        {
            u= 0.01;   // try (nearly) pure Newton every 2n^2 steps
                       // should have good approx to H by then
        }

    } // end-while


    if( !bConverged )
        nLoops= -nLoops;
}




static void gradF( int n, float x[], float F0, float g[] )
{
    int i;
    float a, t, e= G_EPS;

    for ( i=1 ; i<=n ; i++ )
    {
        a= fabs( x[i] );
        if ( a< 1.0 )  a= 1.0;
        a= a*e;
        t= x[i];
        x[i]= t+a;
        g[i]= ( F(n,x) -F0 ) / a;
        x[i]= t;
    }
}





